# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'EyeLinkIO'
copyright = '2024, EyeLinkIO Developers'
author = 'EyeLinkIO Developers'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["numpydoc",
              "sphinx.ext.autodoc",
              "sphinx.ext.intersphinx",
              "sphinx.ext.todo",
              "sphinxemoji.sphinxemoji",
              ]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'pydata_sphinx_theme'
html_static_path = ['_static']

html_css_files = [
    'css/custom.css',
]

html_title = "EyeLinkIO 👀"

# -- Intersphinx configuration -----------------------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "mne": ("https://mne.tools/dev", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
}



# NumPyDoc configuration -----------------------------------------------------

numpydoc_class_members_toctree = False

numpydoc_show_inherited_class_members = False

numpydoc_attributes_as_param_list = True
numpydoc_xref_param_type = True

numpydoc_validate = True

numpydoc_xref_aliases = {
    # Python
    "file-like": ":term:`file-like <python:file object>`",
    "iterator": ":term:`iterator <python:iterator>`",
    "path-like": ":term:`path-like`",
    "array-like": ":term:`array_like <numpy:array_like>`",
    "Path": ":class:`python:pathlib.Path`",
    "bool": ":ref:`bool <python:typebool>`",
    # MNE
    "Raw": "mne.io.Raw",
    "RawEyelink": "mne.io.Raw",
    "Calibration": "mne.preprocessing.eyetracking.Calibration",
    # Pandas
    "DataFrame": "pandas.DataFrame",
    # eyelinkio
    "EDF": ":class:`~eyelinkio.io.EDF`",
}

numpydoc_xref_ignore = {"of", "A",}



from datetime import timedelta, timezone
from pathlib import Path
from warnings import warn

import numpy as np

from ..edf import _defines
from .check import _check_mne_installed, _check_pandas_installed


def _get_test_fnames():
    """Get usable test files (omit EDF if no edf2asc)."""
    path = Path(__file__).parent.parent / "tests" / "data"
    fnames = sorted(list(path.glob("*.edf")))  # test_2.edf will be first
    assert fnames[0].exists()
    return fnames


def to_pandas(edf_obj):
    """Convert an EDF file to a pandas DataFrame.

    Parameters
    ----------
    edf_obj : :class:`EDF`
        The EDF object to convert to a DataFrame.

    Returns
    -------
    df_samples : dict
        A dictionary of :class:`pandas.DataFrame`, containing the samples, blinks,
        saccades, fixations, messages, and calibrations.
    """
    pd = _check_pandas_installed(strict=True)

    dfs = {}
    dfs["discrete"] = {}

    # Discrete
    for key in edf_obj["discrete"]:
        cols = edf_obj["discrete"][key].dtype.names
        dfs["discrete"][key] = pd.DataFrame(edf_obj["discrete"][key], columns=cols)
        # XXX: pyeparse represented messages as byte strings. Should we change that?
        if key == "messages":
            dfs["discrete"][key]["msg"] = dfs["discrete"][key]["msg"].astype(str)
        elif key in ["blinks", "saccades", "fixations"]:
            dfs["discrete"][key]["eye"] = (dfs["discrete"][key]["eye"]).map(
                _defines.eye_constants
            )
    # Samples
    cols = edf_obj["info"]["sample_fields"]
    dfs["samples"] = pd.DataFrame(edf_obj["samples"].T, columns=cols)
    # Calibration
    dfs["calibrations"] = pd.DataFrame(
        edf_obj["info"]["calibrations"].squeeze(),
        columns=edf_obj["info"]["calibrations"].dtype.names,
    )
    return dfs


def to_mne(edf_obj):
    """Create and Return an instance of MNE RawEyelink.

    Parameters
    ----------
    edf_obj : EDF
        The EDF object to convert to an MNE object.

    Returns
    -------
    raw : :class:`mne.io.Raw`
        The MNE RawEyelink instance.
    calibrations : list
        A list of Calibration objects.
    """
    mne = _check_mne_installed(strict=True)

    # in mne we need to specify the eye in the ch name, or pick functions will fail
    eye = edf_obj["info"]["eye"].split("_")[0].lower()
    ch_names = edf_obj["info"]["sample_fields"]
    ch_names = [f"{ch}_{eye}" for ch in ch_names]
    ch_types = []
    more_info = {}
    # Set channel types
    for ch in ch_names:
        if ch.startswith(("xpos", "ypos")):
            ch_types.append("eyegaze")
            if ch.startswith("x"):
                more_info[f"{ch}"] = ("eyegaze", "px", f"{eye}", "x")
            elif ch.startswith("y"):
                more_info[f"{ch}"] = ("eyegaze", "px", f"{eye}", "y")
        elif ch.startswith("ps"):
            ch_types.append("pupil")
            more_info[f"{ch}"] = ("pupil", "au", f"{eye}")
        else:
            warn(f"Unknown channel type: {ch}. Setting to misc.")
            ch_types.append("misc")

    # Create the info structure
    info = mne.create_info(
        ch_names=ch_names, sfreq=edf_obj["info"]["sfreq"], ch_types=ch_types
    )
    # force timezone to UTC
    dt = edf_obj["info"]["meas_date"]
    tz = timezone(timedelta(hours=0))
    dt = dt.replace(tzinfo=tz)
    info.set_meas_date(dt)

    # Create the raw object
    raw = mne.io.RawArray(edf_obj["samples"], info)
    # This will set the loc array etc.
    mne.preprocessing.eyetracking.set_channel_types_eyetrack(raw, more_info)
    # Add annotations
    raw = _add_annotations(edf_obj, raw)
    # Add calibration
    calibrations = _create_calibration(edf_obj)
    return raw, calibrations


def _add_annotations(edf, raw):
    """Add MNE Annotations of EyeLink Events to raw."""
    EYE_EVENTS = [
        ("blinks", "BAD_blink"),
        ("saccades", "saccade"),
        ("fixations", "fixation"),
    ]
    # blinks, saccades, fixations
    for ev, desc in EYE_EVENTS:
        onset = edf["discrete"][ev]["stime"]
        duration = edf["discrete"][ev]["etime"] - onset
        ch_names = [raw.info["ch_names"]] * len(onset)
        raw.annotations.append(onset, duration, desc, ch_names)
    # messages
    onset = edf["discrete"]["messages"]["stime"]
    duration = np.zeros_like(onset)
    desc = edf["discrete"]["messages"]["msg"].astype(str).squeeze()
    raw.annotations.append(onset, duration, desc)
    # TODO: buttons and inputs ?
    return raw


def _create_calibration(edf):
    """Create a calibration event."""
    from mne.preprocessing.eyetracking import Calibration

    calibrations = []
    for ii, this_cal in enumerate(edf["info"]["calibrations"]):
        eye = edf["info"]["eye"].split("_")[0].lower()
        x = this_cal["point_x"]
        y = this_cal["point_y"]
        positions = np.array([x, y]).T
        gx = x + this_cal["diff_x"]
        gy = y + this_cal["diff_y"]
        gaze = np.array([gx, gy]).T
        offsets = this_cal["offset"]
        avg_error = np.mean(this_cal["offset"])
        max_error = np.max(this_cal["offset"])
        screen_resolution = edf["info"]["screen_coords"]
        # XXX: getting onset and model will be tricky.
        # XXX: for binocular data, we will need to get the eye another way
        cal = Calibration(
            onset=None,
            model=None,
            eye=eye,
            avg_error=avg_error,
            max_error=max_error,
            positions=positions,
            offsets=offsets,
            gaze=gaze,
            screen_resolution=screen_resolution,
        )
        calibrations.append(cal)
    return calibrations



# Authors: Scott Huberty <seh33@uw.edu>
# Aknowledgements to the devs of the pyeparse package, from which this is derived
#
# License: BSD (3-clause)

from . import utils

from .edf import read_edf, EDF

try:
    from importlib.metadata import version

    __version__ = version("eyelinkio")
except Exception:
    __version__ = "0.0.0"


"""
Defines for the EDF API.

This file contains all the define values from the edfapi header files.
A couple extra have been added as well.

Some of the constants are grouped into dicts, where both k,v and v,k entries
are added so it is easy to go from either the int -> str constant or str -> int
constant value for each.

The constants that are not grouped into sub dicts are available either as
variables within the defines module, or in dict form (named 'edf_constants')
where both k,v and v,k entries are added so it is easy to go from either the
int -> str constant or str -> int constant value for each.

Created on Sat Mar 15 09:40:17 2014

@author: Sol
"""


def create_constants(var_val_dict):
    # define function vars here.
    # They will be read and a constants dict with var_name:var_val
    # and var_val:var_name will be created abd returned

    constants = dict(
        [
            (key, val)
            for key, val in dict(var_val_dict).items()
            if (not key.startswith("_") and not callable(val))
        ]
    )
    constants.update([(val, key) for key, val in dict(constants).items()])
    return constants


# ************* EVENT TYPE CODES ***********
event_constants = create_constants(
    dict(
        SAMPLE_TYPE=200,
        # buffer = IEVENT, FEVENT, btype = IEVENT_BUFFER
        STARTPARSE=1,  # these only have time and eye data
        ENDPARSE=2,
        BREAKPARSE=10,
        # EYE DATA: contents determined by evt_data
        STARTBLINK=3,  # and by "read" data item
        ENDBLINK=4,  # all use IEVENT format
        STARTSACC=5,
        ENDSACC=6,
        STARTFIX=7,
        ENDFIX=8,
        FIXUPDATE=9,
        # buffer = (none, directly affects state), btype = CONTROL_BUFFER
        # control events: all put data into
        # the EDF_FILE or ILINKDATA status
        STARTSAMPLES=15,  # start of events in block
        ENDSAMPLES=16,  # end of samples in block
        STARTEVENTS=17,  # start of events in block
        ENDEVENTS=18,  # end of events in block
        # buffer = IMESSAGE, btype = IMESSAGE_BUFFER
        MESSAGEEVENT=24,  # user-definable text or data
        # buffer = IOEVENT, btype = IOEVENT_BUFFER
        BUTTONEVENT=25,  # button state change
        INPUTEVENT=28,  # change of input port
        LOST_DATA_EVENT=0x3F,  # NEW: Event flags gap in data stream
        NO_PENDING_ITEMS=0,
        RECORDING_INFO=30,
    )
)

# Missing data constants
MISSING_DATA = -32768  # data is missing (integer)
MISSING = -32768
INaN = -32768

eye_constants = create_constants(
    dict(
        # binocular data needs to ID the eye for events
        # samples need to index the data
        # These constants are used as eye identifiers
        LEFT_EYE=0,  # index and ID of eyes
        RIGHT_EYE=1,
        BINOCULAR=2,  # data for both eyes available
    )
)


pupil_constants = create_constants(dict(PUPIL_AREA=0, PUPIL_DIAMETER=1))

# ******** EYE SAMPLE DATA FORMATS ******
#
# The SAMPLE struct contains data from one 4-msec
# eye-tracker sample. The <flags> field has a bit for each
# type of data in the sample. Fields not read have 0 flag
# bits, and are set to MISSING_DATA
#
# flags to define what data is included in each sample.
# There is one bit for each type.  Total data for samples
# in a block is indicated by these bits in the <sam_data>
# field of ILINKDATA or EDF_FILE, and is updated by the
# STARTSAMPLES control event.

SAMPLE_LEFT = 0x8000  # data for these eye(s)
SAMPLE_RIGHT = 0x4000
SAMPLE_TIMESTAMP = 0x2000  # always for link, used to compress files
SAMPLE_PUPILXY = 0x1000  # pupil x,y pair
SAMPLE_HREFXY = 0x0800  # head-referenced x,y pair
SAMPLE_GAZEXY = 0x0400  # gaze x,y pair
SAMPLE_GAZERES = 0x0200  # gaze res (x,y pixels per degree) pair
SAMPLE_PUPILSIZE = 0x0100  # pupil size
SAMPLE_STATUS = 0x0080  # error flags
SAMPLE_INPUTS = 0x0040  # input data port
SAMPLE_BUTTONS = 0x0020  # button state: LSBy state, MSBy changes
SAMPLE_HEADPOS = 0x0010  # head-position: byte tells # words
SAMPLE_TAGGED = 0x0008  # reserved variable-length tagged
SAMPLE_UTAGGED = 0x0004  # user-defineabe variable-length tagged
SAMPLE_ADD_OFFSET = 0x0002  # if this flag is set for the sample add .5ms to
# the sample time

# ************ CONSTANTS FOR EVENTS ***********
# "read" flag contents in IEVENT
# time data
READ_ENDTIME = 0x0040  # end time (start time always read)

# non-position eye data:
READ_GRES = 0x0200  # gaze resolution xy
READ_SIZE = 0x0080  # pupil size
READ_VEL = 0x0100  # velocity (avg, peak)
READ_STATUS = 0x2000  # status (error word)
READ_BEG = 0x0001  # event has start data for vel,size,gres
READ_END = 0x0002  # event has end data for vel,size,gres
READ_AVG = 0x0004  # event has avg pupil size, velocity

# position eye data
READ_PUPILXY = 0x0400  # pupilxy REPLACES gaze, href data if read
READ_HREFXY = 0x0800
READ_GAZEXY = 0x1000
READ_BEGPOS = 0x0008  # position data for these parts of event
READ_ENDPOS = 0x0010
READ_AVGPOS = 0x0020

# RAW FILE/LINK CODES: REVERSE IN R/W
FRIGHTEYE_EVENTS = 0x8000  # has right eye events
FLEFTEYE_EVENTS = 0x4000  # has left eye events

# "event_types" flag in ILINKDATA or EDF_FILE
# tells what types of events were written by tracker
LEFTEYE_EVENTS = 0x8000  # has left eye events
RIGHTEYE_EVENTS = 0x4000  # has right eye events
BLINK_EVENTS = 0x2000  # has blink events
FIXATION_EVENTS = 0x1000  # has fixation events
FIXUPDATE_EVENTS = 0x0800  # has fixation updates
SACCADE_EVENTS = 0x0400  # has saccade events
MESSAGE_EVENTS = 0x0200  # has message events
BUTTON_EVENTS = 0x0040  # has button events
INPUT_EVENTS = 0x0020  # has input port events

# "event_data" flags in ILINKDATA or EDF_FILE
# tells what types of data were included in events by tracker
EVENT_VELOCITY = 0x8000  # has velocity data
EVENT_PUPILSIZE = 0x4000  # has pupil size data
EVENT_GAZERES = 0x2000  # has gaze resolution
EVENT_STATUS = 0x1000  # has status flags
EVENT_GAZEXY = 0x0400  # has gaze xy position
EVENT_HREFXY = 0x0200  # has head-ref xy position
EVENT_PUPILXY = 0x0100  # has pupil xy position
FIX_AVG_ONLY = 0x0008  # only avg. data to fixation evts
START_TIME_ONLY = 0x0004  # only start-time in start events
PARSEDBY_GAZE = 0x00C0  # how events were generated
PARSEDBY_HREF = 0x0080
PARSEDBY_PUPIL = 0x0040

# *********** STATUS FLAGS (samples and events) ***************
LED_TOP_WARNING = 0x0080  # marker is in border of image
LED_BOT_WARNING = 0x0040
LED_LEFT_WARNING = 0x0020
LED_RIGHT_WARNING = 0x0010
HEAD_POSITION_WARNING = 0x00F0  # head too far from calibr???
LED_EXTRA_WARNING = 0x0008  # glitch or extra markers
LED_MISSING_WARNING = 0x0004  # <2 good data points in last 100 msec)
HEAD_VELOCITY_WARNING = 0x0001  # head moving too fast
CALIBRATION_AREA_WARNING = 0x0002  # pupil out of good mapping area
MATH_ERROR_WARNING = 0x2000  # math error in proc. sample

# THESE CODES ONLY VALID FOR EYELINK II

# this sample interpolated to preserve sample rate
# usually because speed dropped due to missing pupil
# INTERP_SAMPLE_WARNING 0x1000

# pupil interpolated this sample
# usually means pupil loss or
# 500 Hz sample with CR but no pupil
#
INTERP_PUPIL_WARNING = 0x8000

# all CR-related errors
CR_WARNING = 0x0F00
CR_LEFT_WARNING = 0x0500
CR_RIGHT_WARNING = 0x0A00

# CR is actually lost
CR_LOST_WARNING = 0x0300
CR_LOST_LEFT_WARNING = 0x0100
CR_LOST_RIGHT_WARNING = 0x0200

# this sample has interpolated/held CR
CR_RECOV_WARNING = 0x0C00
CR_RECOV_LEFT_WARNING = 0x0400
CR_RECOV_RIGHT_WARNING = 0x0800

TFLAG_MISSING = 0x4000  # missing
TFLAG_ANGLE = 0x2000  # extreme target angle
TFLAG_NEAREYE = 0x1000  # target near eye so windows overlapping
# DISTANCE WARNINGS (limits set by remote_distance_warn_range command)
TFLAG_CLOSE = 0x0800  # distance vs. limits
TFLAG_FAR = 0x0400
# TARGET TO CAMERA EDGE  (margin set by remote_edge_warn_pixels command)
TFLAG_T_TSIDE = 0x0080  # target near edge of image (left, right, top, bottom)
TFLAG_T_BSIDE = 0x0040
TFLAG_T_LSIDE = 0x0020
TFLAG_T_RSIDE = 0x0010
# EYE TO CAMERA EDGE  (margin set by remote_edge_warn_pixels command)
TFLAG_E_TSIDE = 0x0008  # eye near edge of image (left, right, top, bottom)
TFLAG_E_BSIDE = 0x0004
TFLAG_E_LSIDE = 0x0002
TFLAG_E_RSIDE = 0x0001

PUPIL_ONLY_250 = 0
PUPIL_ONLY_500 = 1
PUPIL_CR = 2

EYELINK = 1
EYELINK_II = 2
EYELINK_1000 = 3

edf_constants = dict(
    [
        (k, v)
        for k, v in dict(locals()).items()
        if (not k.startswith("_") and not isinstance(v, dict))
    ]
)



from importlib import import_module


def _has_edfapi():
    """Determine if a user has edfapi installed."""
    from ..edf.read_edf import has_edfapi
    return has_edfapi


def _check_edfapi():
    """Check if edfapi is installed."""
    if not _has_edfapi():
        raise RuntimeError('edfapi is not installed')


def requires_edfapi(func):
    """Skip testing if edfapi is not installed."""
    import pytest

    return pytest.mark.skipif(not _has_edfapi(), reason='Requires edfapi')(func)


def _check_pandas_installed(strict=True):
    """Aux function."""
    return _soft_import("pandas", "converting to dataframes", strict=strict)

def _check_mne_installed(strict=True):
    """Aux function."""
    return _soft_import("mne", "exporting to MNE", strict=strict)


def _soft_import(name, purpose, strict=True):
    """Import soft dependencies, providing informative errors on failure.

    Parameters
    ----------
    name : str
        Name of the module to be imported. For example, 'pandas'.
    purpose : str
        A very brief statement (formulated as a noun phrase) explaining what
        functionality the package provides to MNE-Python.
    strict : bool
        Whether to raise an error if module import fails.
    """

    # so that error msg lines are aligned
    def indent(x):
        return x.rjust(len(x) + 14)

    try:
        mod = import_module(name)
        return mod
    except (ImportError, ModuleNotFoundError):
        if strict:
            raise RuntimeError(
                f"For {purpose} to work, the {name} module is needed, "
                + "but it could not be imported.\n"
                + "\n".join(
                    (
                        indent(
                            "use the following installation method "
                            "appropriate for your environment:"
                        ),
                        indent(f"'pip install {name}'"),
                        indent(f"'conda install -c conda-forge {name}'"),
                    )
                )
            )
        else:
            return False



"""Functions for reading Eyelink EDF Files."""

import ctypes as ct
import warnings
from datetime import datetime
from functools import partial
from os import path as op
from pathlib import Path

import numpy as np

try:
    from ._edf2py import (
        edf_close_file,
        edf_get_float_data,
        edf_get_next_data,
        edf_get_preamble_text,
        edf_get_preamble_text_length,
        edf_get_version,
        edf_open_file,
    )

    has_edfapi = True
    why_not = None
except OSError as exp:
    (
        edf_open_file,
        edf_close_file,
        edf_get_next_data,
        edf_get_preamble_text,
        edf_get_preamble_text_length,
        edf_get_version,
        edf_get_float_data,
    ) = [None] * 8
    has_edfapi = False
    why_not = str(exp)

from . import _defines as defines
from ._defines import event_constants

_MAX_MSG_LEN = 260  # maxmimum message length we'll need to store


def read_edf(fname):
    """Read an EyeLink EDF file.

    Parameters
    ----------
    fname : path-like
        The name of the EDF file.

    Returns
    -------
    edf : EDF
        An instance of EDF:  The EyeLink data represented in Python.
    """
    return EDF(fname)


class EDF(dict):
    """Represent EyeLink EDF files in Python.

    This class is a subclass of :class:`dict`, and so can be indexed like a
    dictionary. To see the available keys, use the ``keys()`` method.

    Parameters
    ----------
    fname : str
        The name of the EDF file.
    """

    def __init__(self, fname):
        if not has_edfapi:
            raise OSError("Could not load EDF api: %s" % why_not)
        info, discrete, times, samples = _read_raw_edf(fname)
        self.info = info
        self.info["filename"] = Path(fname).name
        self.discrete = discrete
        self._times = times
        self._samples = samples
        super().__init__(
            info=info, discrete=discrete, times=times, samples=samples
        )

    def __repr__(self):
        """Return a summary of the EDF File."""
        return (
            f"<EDF | {self['info']['filename']}> \n"
            f"  Version: {self['info']['version']} \n"
            f"  Eye: {self['info']['eye']} \n"
            f"  Pupil unit: {self['info']['ps_units']} \n"
            f"  Sampling frequency: {self['info']['sfreq']} Hz \n"
            f"  Calibrations: {len(self['info']['calibrations'])} \n"
            f"  Length: {len(self['times']) / self['info']['sfreq']} seconds \n"
        )

    def to_pandas(self):
        """Convert an EDF file to a pandas DataFrame.

        Returns
        -------
        df_samples : dict of DataFrame
            A dictionary of :class:`~pandas.DataFrame`'s, containing the samples,
            blinks, saccades, fixations, messages, and calibrations.
        """
        from ..utils import to_pandas
        return to_pandas(self)

    def to_mne(self):
        """Create an MNE Raw object from the EDF object.

        Returns
        -------
        raw : RawEyelink
            An instance of Raw.
        calibrations : list of Calibration
            A list of Calibration objects.
        """
        from ..utils import to_mne
        return to_mne(self)

class _edf_open:
    """Context manager for opening EDF files."""

    def __init__(self, fname):
        self.fname = op.normpath(op.abspath(fname).encode("ASCII"))
        self.fid = None

    def __enter__(self):
        error_code = ct.c_int(1)
        self.fid = edf_open_file(self.fname, 2, 1, 1, ct.byref(error_code))
        if self.fid is None or error_code.value != 0:
            raise OSError(
                'Could not open file "%s": (%s, %s)'
                % (self.fname, self.fid, error_code.value)
            )
        return self.fid

    def __exit__(self, type_, value, traceback):
        if self.fid is not None:
            result = edf_close_file(self.fid)
            if result != 0:
                raise OSError('File "%s" could not be closed' % self.fname)


_ets2pp = dict(
    SAMPLE_TYPE="sample",
    ENDFIX="fixations",
    ENDSACC="saccades",
    ENDBLINK="blinks",
    BUTTONEVENT="buttons",
    INPUTEVENT="inputs",
    MESSAGEEVENT="messages",
)


def _read_raw_edf(fname):
    """Read data from raw EDF file into pyeparse format."""
    if not op.isfile(fname):
        raise OSError('File "%s" does not exist' % fname)

    #
    # First pass: get the number of each type of sample
    #
    n_samps = dict()
    offsets = dict()
    for key in _ets2pp.values():
        n_samps[key] = 0
        offsets[key] = 0
    with _edf_open(fname) as edf:
        etype = None
        while etype != event_constants.get("NO_PENDING_ITEMS"):
            etype = edf_get_next_data(edf)
            if etype not in event_constants:
                raise RuntimeError("unknown type %s" % event_constants[etype])
            ets = event_constants[etype]
            if ets in _ets2pp:
                n_samps[_ets2pp[ets]] += 1

    #
    # Now let's actually read in the data
    #
    with _edf_open(fname) as edf:
        info = _parse_preamble(edf)
        etype = None
        res = dict(
            info=info,
            samples=None,
            n_samps=n_samps,
            offsets=offsets,
            edf_fields=dict(messages=["stime", "msg"]),
            discrete=dict(),
        )
        # XXX: pyeparse represented messages as byte strings.
        # XXX: Maybe we should use regular python strings?
        dtype = [("stime", np.float64), ("msg", "|S%s" % _MAX_MSG_LEN)]
        res["discrete"]["messages"] = np.empty((n_samps["messages"]), dtype=dtype)
        res["eye_idx"] = None  # in case we get input/button before START
        while etype != event_constants.get("NO_PENDING_ITEMS"):
            etype = edf_get_next_data(edf)
            if etype not in event_constants:
                raise RuntimeError("unknown type %s" % event_constants[etype])
            ets = event_constants[etype]
            _element_handlers[ets](edf, res)
        _element_handlers["VERSION"](res)

    #
    # Put info and discrete into correct output format
    #
    discrete = res["discrete"]
    info = res["info"]
    event_types = ("saccades", "fixations", "blinks", "buttons", "inputs", "messages")
    info["sample_fields"] = info["sample_fields"][1:]  # omit time

    #
    # fix sample times
    #
    data = res["samples"][1:]
    data[data >= 100000000.0 - 1] = np.nan
    orig_times = res["samples"][0]  # original times
    assert np.array_equal(orig_times, np.sort(orig_times))
    times = np.arange(len(orig_times), dtype=np.float64) / info["sfreq"]
    for key in event_types:
        for sub_key in ("stime", "etime"):
            if sub_key in discrete[key].dtype.names:
                _adjust_time(discrete[key][sub_key], orig_times, times)

    _extract_calibration(info, discrete["messages"])

    # now we correct our time offsets
    return info, discrete, times, data


def _adjust_time(x, orig_times, times):
    """Adjust time, inplace."""
    x[:] = np.interp(x, orig_times, times)


def _extract_calibration(info, messages):
    """Extract calibration from messages."""
    lines = []
    for msg in messages["msg"]:
        msg = msg.decode("ASCII")
        if msg.startswith("!CAL") or msg.startswith("VALIDATE"):
            lines.append(msg)
        if msg.startswith("GAZE_COORDS"):
            coords = msg.split()[-4:]
            coords = [int(round(float(c))) for c in coords]
            info["screen_coords"] = np.array(
                [coords[2] - coords[0] + 1, coords[3] - coords[1] + 1], int
            )
    calibrations = list()
    keys = ["point_x", "point_y", "offset", "diff_x", "diff_y"]
    li = 0
    while li < len(lines):
        line = lines[li]
        if "!CAL VALIDATION " in line and "ABORTED" not in line:
            cal_kind = line.split("!CAL VALIDATION ")[1].split()[0]
            n_points = int([c for c in cal_kind if c.isdigit()][0])
            this_validation = []
            for ni in range(n_points):
                subline = lines[li + ni + 1].split()
                xy = subline[-6].split(",")
                xy_diff = subline[-2].split(",")
                vals = [
                    float(v)
                    for v in [xy[0], xy[1], subline[-4], xy_diff[0], xy_diff[1]]
                ]
                this_validation.append(vals)
            li += n_points
            this_validation = np.array(this_validation)
            dtype = [(key, "f8") for key in keys]
            out = np.empty(len(this_validation), dtype=dtype)
            for key, data in zip(keys, this_validation.T):
                out[key] = data
            calibrations.append(out)
        li += 1
    info["calibrations"] = np.array(calibrations)


def _extract_sys_info(line):
    """Aux function for preprocessing sys info lines."""
    return line[line.find(":") :].strip(": \r\n")


def _parse_preamble(edf):
    tlen = edf_get_preamble_text_length(edf)
    txt = ct.create_string_buffer(tlen)
    edf_get_preamble_text(edf, txt, tlen + 1)
    preamble_lines = txt.value.decode("ASCII").split("\n")
    info = dict()
    for line in preamble_lines:
        if "!MODE" in line:
            line = line.split()
            info["eye"], info["sfreq"] = line[-1], float(line[-4])
        elif "DATE:" in line:
            line = _extract_sys_info(line).strip()
            fmt = "%a %b  %d %H:%M:%S %Y"
            info["meas_date"] = datetime.strptime(line, fmt)
        elif "VERSION:" in line:
            info["version"] = _extract_sys_info(line)
        elif "CAMERA:" in line:
            info["camera"] = _extract_sys_info(line)
        elif "SERIAL NUMBER:" in line:
            info["serial"] = _extract_sys_info(line)
        elif "CAMERA_CONFIG:" in line:
            info["camera_config"] = _extract_sys_info(line)
    return info


def _to_list(element, keys, idx):
    """Return a list of particular fields of an EyeLink data element."""
    out = list()
    for k in keys:
        v = getattr(element, k)
        if hasattr(v, "_length_"):
            out.append(v[idx])
        else:
            out.append(v)
    return out


def _sample_fields_available(sflags):
    """Indicate which fields are available in a sample.

    Returns a dict where the keys indicate fields (or field groups) of a
    sample; the value for each indicates if the field has been populated
    with data and can be considered as useful information.
    """
    return dict(
        time=bool(sflags & defines.SAMPLE_TIMESTAMP),  # sample time
        gx=bool(sflags & defines.SAMPLE_GAZEXY),  # gaze X position
        gy=bool(sflags & defines.SAMPLE_GAZEXY),  # gaze Y position
        pa=bool(sflags & defines.SAMPLE_PUPILSIZE),  # pupil size
        left=bool(sflags & defines.SAMPLE_LEFT),  # left eye data
        right=bool(sflags & defines.SAMPLE_RIGHT),  # right eye data
        pupilxy=bool(sflags & defines.SAMPLE_PUPILXY),  # raw eye position
        hrefxy=bool(sflags & defines.SAMPLE_HREFXY),  # href eye position
        gazeres=bool(sflags & defines.SAMPLE_GAZERES),  # x,y pixels per deg
        status=bool(sflags & defines.SAMPLE_STATUS),  # sample status
        inputs=bool(sflags & defines.SAMPLE_INPUTS),  # sample inputs
        button=bool(sflags & defines.SAMPLE_BUTTONS),  # sample buttons
        headpos=bool(sflags & defines.SAMPLE_HEADPOS),  # sample head pos
        # if this flag is set for the sample add .5ms to the sample time
        addoffset=bool(sflags & defines.SAMPLE_ADD_OFFSET),
        # reserved variable-length tagged
        tagged=bool(sflags & defines.SAMPLE_TAGGED),
        # user-defineabe variable-length tagged
        utagged=bool(sflags & defines.SAMPLE_UTAGGED),
    )


'''
def _event_fields_available(eflags):
    """
    Returns a dict where the keys indicate fields (or field groups) of an
    EDF event; the value for each indicates if the field has been populated
    with data and can be considered as useful information.
    """
    return dict(
        endtime=bool(eflags & defines.READ_ENDTIME),  # end time
        gres=bool(eflags & defines.READ_GRES),  # gaze resolution xy
        size=bool(eflags & defines.READ_SIZE),  # pupil size
        vel=bool(eflags & defines.READ_VEL),  # velocity (avg, peak)
        status=bool(eflags & defines.READ_STATUS),  # status (error word)
        beg=bool(eflags & defines.READ_BEG),  # start data for vel,size,gres
        end=bool(eflags & defines.READ_END),  # end data for vel,size,gres
        avg=bool(eflags & defines.READ_AVG),  # avg pupil size, velocity
        pupilxy=bool(eflags & defines.READ_PUPILXY),  # position eye data
        hrefxy=bool(eflags & defines.READ_HREFXY),
        gazexy=bool(eflags & defines.READ_GAZEXY),
        begpos=bool(eflags & defines.READ_BEGPOS),
        endpos=bool(eflags & defines.READ_ENDPOS),
        avgpos=bool(eflags & defines.READ_AVGPOS),
    )
'''


_pp2el = dict(
    eye="eye",
    time="time",
    stime="sttime",
    etime="entime",
    xpos="gx",
    ypos="gy",
    sxp="gstx",
    syp="gsty",
    exp="genx",
    eyp="geny",
    axp="gavx",
    ayp="gavy",
    pv="pvel",
    ps="pa",
    aps="avg",
    buttons="buttons",
    input="input",
)
_el2pp = dict()
for key, val in _pp2el.items():
    _el2pp[val] = key


#
# EDF File Handlers
#


def _handle_recording_info(edf, res):
    """RECORDING_INFO."""
    info = res["info"]
    e = edf_get_float_data(edf).contents.rec
    if e.state == 0:  # recording stopped
        return
    if "sfreq" in info:
        assert e.sample_rate == info["sfreq"]
        assert defines.eye_constants[e.eye -1] == info["eye"]
        x = str(defines.pupil_constants[e.pupil_type])
        assert x == info["ps_units"]
        return
    info["sfreq"] = e.sample_rate
    info["ps_units"] = defines.pupil_constants[e.pupil_type]
    # TODO: edfapi eye constants are 1-based, ours are 0-based. Fix this in _defines?
    info["eye"] = defines.eye_constants[e.eye -1]
    res["eye_idx"] = e.eye - 1 # This should be 0: left, 1: right, 2: binocular
    if res["eye_idx"] == 2:
        raise NotImplementedError("Reading Binocular data is not yet supported.")

    # Figure out sample flags
    sflags = _sample_fields_available(e.sflags)
    edf_fields = ["time", "gx", "gy", "pa"]  # XXX Expand?
    edf_fields = [field for field in edf_fields if sflags[field]]
    sample_fld = [_el2pp[field] for field in edf_fields]
    res["edf_sample_fields"] = edf_fields
    res["info"]["sample_fields"] = sample_fld
    res["samples"] = np.empty((len(edf_fields), res["n_samps"]["sample"]), np.float64)


def _handle_sample(edf, res):
    """SAMPLE_TYPE."""
    e = edf_get_float_data(edf).contents.fs
    off = res["offsets"]["sample"]
    res["samples"][:, off] = _to_list(e, res["edf_sample_fields"], res["eye_idx"])
    res["offsets"]["sample"] += 1


def _handle_message(edf, res):
    """MESSAGEEVENT."""
    e = edf_get_float_data(edf).contents.fe
    msg = ct.string_at(ct.byref(e.message[0]), e.message.contents.len + 1)[2:]
    msg = msg.decode("UTF-8")
    msg = "".join([i if ord(i) < 128 else "" for i in msg])
    if len(msg) > _MAX_MSG_LEN:
        warnings.warn("Message truncated to %s characters:\n%s" % (_MAX_MSG_LEN, msg))
    off = res["offsets"]["messages"]
    res["discrete"]["messages"]["stime"][off] = e.sttime
    res["discrete"]["messages"]["msg"][off] = msg[:_MAX_MSG_LEN]
    res["offsets"]["messages"] += 1


def _handle_end(edf, res, name):
    """ENDSACC, ENDFIX, ENDBLINK, BUTTONS, INPUT."""
    if name not in res["discrete"]:
        # XXX This should be changed to support given fields
        if name == "saccades":
            f = ["eye", "sttime", "entime", "gstx", "gsty", "genx", "geny", "pvel"]
        elif name == "fixations":
            f = ["eye", "sttime", "entime", "gavx", "gavy"]
        elif name == "blinks":
            f = ["eye", "sttime", "entime"]
        elif name == "buttons":
            f = ["sttime", "buttons"]
        elif name == "inputs":
            f = ["sttime", "input"]
        else:
            raise KeyError("Unknown name %s" % name)
        res["edf_fields"][name] = f
        our_names = [_el2pp[field] for field in f]
        dtype = [(ff, np.float64) for ff in our_names]
        res["discrete"][name] = np.empty(res["n_samps"][name], dtype=dtype)
    e = edf_get_float_data(edf).contents.fe
    vals = _to_list(e, res["edf_fields"][name], res["eye_idx"])
    off = res["offsets"][name]
    for ff, vv in zip(res["discrete"][name].dtype.names, vals):
        res["discrete"][name][ff][off] = vv
    res["offsets"][name] += 1

def _handle_pass(edf, res):
    """Events we don't care about or haven't had to care about yet."""
    pass


def _handle_fixation_update(edf, res):
    """FIXUPDATE."""
    raise NotImplementedError


def _handle_version(res):
    """EDFAPI VERSION."""
    version = edf_get_version()
    res["info"]["edfapi_version"] = version.decode("utf-8")


# element_handlers maps the various EDF file element types to the
# element handler function that should be called.

_element_handlers = dict(
    RECORDING_INFO=_handle_recording_info,
    SAMPLE_TYPE=_handle_sample,
    MESSAGEEVENT=_handle_message,
    ENDFIX=partial(_handle_end, name="fixations"),
    ENDSACC=partial(_handle_end, name="saccades"),
    ENDBLINK=partial(_handle_end, name="blinks"),
    BUTTONEVENT=partial(_handle_end, name="buttons"),
    INPUTEVENT=partial(_handle_end, name="inputs"),
    STARTFIX=_handle_pass,
    STARTSACC=_handle_pass,
    STARTBLINK=_handle_pass,
    STARTPARSE=_handle_pass,
    FIXUPDATE=_handle_pass,
    ENDPARSE=_handle_pass,
    NO_PENDING_ITEMS=_handle_pass,  # context manager
    BREAKPARSE=_handle_pass,
    STARTSAMPLES=_handle_pass,
    ENDSAMPLES=_handle_pass,
    STARTEVENTS=_handle_pass,
    ENDEVENTS=_handle_pass,
    VERSION=_handle_version,
)



"""Wrapper for libedfapi.so."""

import struct
import sys
from ctypes import (
    CDLL,
    POINTER,
    Structure,
    Union,
    c_char,
    c_char_p,
    c_float,
    c_int,
    c_short,
    c_ubyte,
    c_uint,
    c_ushort,
    util,
)

# find and load the library
if sys.platform.startswith('win') and (8 * struct.calcsize("P") == 64):
    name = 'edfapi64'  # on windows you can install both
else:
    name = 'edfapi'
fname = util.find_library(name)
if fname is None:
    raise OSError('edfapi not found')
edfapi = CDLL(fname)


# Extract the functions we need
class LSTRING(Structure):
    pass


LSTRING.__slots__ = ['len', 'c']
LSTRING._fields_ = [('len', c_short), ('c', c_char)]


class FSAMPLE(Structure):
    pass


cf = c_float
cf2 = c_float * 2
FSAMPLE.__slots__ = [
    'time', 'px', 'py', 'hx', 'hy', 'pa', 'gx', 'gy', 'rx', 'ry',
    'gxvel', 'gyvel', 'hxvel', 'hyvel', 'rxvel', 'ryvel', 'fgxvel', 'fgyvel',
    'fhxvel', 'fhyvel', 'frxvel', 'fryvel', 'hdata', 'flags', 'input',
    'buttons', 'htype', 'errors']
FSAMPLE._fields_ = [
    ('time', c_uint), ('px', cf2), ('py', cf2), ('hx', cf2),
    ('hy', cf2), ('pa', cf2), ('gx', cf2), ('gy', cf2),
    ('rx', cf), ('ry', cf), ('gxvel', cf2), ('gyvel', cf2),
    ('hxvel', cf2), ('hyvel', cf2), ('rxvel', cf2), ('ryvel', cf2),
    ('fgxvel', cf2), ('fgyvel', cf2), ('fhxvel', cf2), ('fhyvel', cf2),
    ('frxvel', cf2), ('fryvel', cf2), ('hdata', c_short * 8),
    ('flags', c_ushort), ('input', c_ushort), ('buttons', c_ushort),
    ('htype', c_short), ('errors', c_ushort)]


class FEVENT(Structure):
    pass


FEVENT.__slots__ = [
    'time', 'type', 'read', 'sttime', 'entime', 'hstx', 'hsty', 'gstx', 'gsty',
    'sta', 'henx', 'heny', 'genx', 'geny', 'ena', 'havx', 'havy',
    'gavx', 'gavy', 'ava', 'avel', 'pvel', 'svel', 'evel', 'supd_x',
    'eupd_x', 'supd_y', 'eupd_y', 'eye', 'status', 'flags', 'input', 'buttons',
    'parsedby', 'message']
FEVENT._fields_ = [
    ('time', c_uint), ('type', c_short), ('read', c_ushort),
    ('sttime', c_uint), ('entime', c_uint), ('hstx', cf), ('hsty', cf),
    ('gstx', cf), ('gsty', cf), ('sta', cf), ('henx', cf), ('heny', cf),
    ('genx', cf), ('geny', cf), ('ena', cf), ('havx', cf), ('havy', cf),
    ('gavx', cf), ('gavy', cf), ('ava', cf), ('avel', cf), ('pvel', cf),
    ('svel', cf), ('evel', cf), ('supd_x', cf), ('eupd_x', cf),
    ('supd_y', cf), ('eupd_y', cf), ('eye', c_short), ('status', c_ushort),
    ('flags', c_ushort), ('input', c_ushort), ('buttons', c_ushort),
    ('parsedby', c_ushort), ('message', POINTER(LSTRING))]


class RECORDINGS(Structure):
    pass


RECORDINGS.__slots__ = [
    'time', 'sample_rate', 'eflags', 'sflags', 'state', 'record_type',
    'pupil_type', 'recording_mode', 'filter_type', 'pos_type', 'eye']
RECORDINGS._fields_ = [
    ('time', c_uint), ('sample_rate', cf), ('eflags', c_ushort),
    ('sflags', c_ushort), ('state', c_ubyte), ('record_type', c_ubyte),
    ('pupil_type', c_ubyte), ('recording_mode', c_ubyte),
    ('filter_type', c_ubyte), ('pos_type', c_ubyte), ('eye', c_ubyte)]


class EDFFILE(Structure):
    pass


edf_open_file = edfapi.edf_open_file
edf_open_file.argtypes = [c_char_p, c_int, c_int, c_int, POINTER(c_int)]
edf_open_file.restype = POINTER(EDFFILE)

edf_close_file = edfapi.edf_close_file
edf_close_file.argtypes = [POINTER(EDFFILE)]
edf_close_file.restype = c_int

edf_get_next_data = edfapi.edf_get_next_data
edf_get_next_data.argtypes = [POINTER(EDFFILE)]
edf_get_next_data.restype = c_int

edf_get_version = edfapi.edf_get_version
edf_get_version.restype = c_char_p


edf_get_preamble_text_length = edfapi.edf_get_preamble_text_length
edf_get_preamble_text_length.argtypes = [POINTER(EDFFILE)]
edf_get_preamble_text_length.restype = c_int

edf_get_preamble_text = edfapi.edf_get_preamble_text
edf_get_preamble_text.argtypes = [POINTER(EDFFILE), c_char_p, c_int]
edf_get_preamble_text.restype = c_int

if hasattr(edfapi, 'edf_get_recording_data'):
    edf_get_recording_data = edfapi.edf_get_recording_data
else:
    # edf_get_recording_data is edf_get_recording on recent Linux versions
    edf_get_recording_data = edfapi.edf_get_recording
edf_get_recording_data.argtypes = [POINTER(EDFFILE)]
edf_get_recording_data.restype = POINTER(RECORDINGS)

if hasattr(edfapi, 'edf_get_sample_data'):
    edf_get_sample_data = edfapi.edf_get_sample_data
else:
    # edf_get_sample_data is edf_get_sample on recent Linux versions
    edf_get_sample_data = edfapi.edf_get_sample
edf_get_sample_data.argtypes = [POINTER(EDFFILE)]
edf_get_sample_data.restype = POINTER(FSAMPLE)

if hasattr(edfapi, 'edf_get_event_data'):
    edf_get_event_data = edfapi.edf_get_event_data
else:
    # edf_get_event_data is edf_get_event on recent Linux versions
    edf_get_event_data = edfapi.edf_get_event
edf_get_event_data.argtypes = [POINTER(EDFFILE)]
edf_get_event_data.restype = POINTER(FEVENT)


class IMESSAGE(Structure):
    pass

IMESSAGE.__slots__ = ['time', 'type', 'length', 'text']
IMESSAGE._fields_ = [('time', c_uint), ('type', c_short), ('length', c_ushort),
                     ('text', c_ubyte * 260)]


class IOEVENT(Structure):
    pass

IOEVENT.__slots__ = ['time', 'type', 'data']
IOEVENT._fields_ = [('time', c_uint), ('type', c_short), ('data', c_ushort)]


class ALLF_DATA(Union):
    pass

ALLF_DATA.__slots__ = ['fe', 'im', 'io', 'fs', 'rec']
ALLF_DATA._fields_ = [('fe', FEVENT), ('im', IMESSAGE), ('io', IOEVENT),
                      ('fs', FSAMPLE), ('rec', RECORDINGS)]

edf_get_float_data = edfapi.edf_get_float_data
edf_get_float_data.argtypes = [POINTER(EDFFILE)]
edf_get_float_data.restype = POINTER(ALLF_DATA)

"""
edf_get_sample_close_to_time = edfapi.edf_get_sample_close_to_time
edf_get_sample_close_to_time.argtypes = [POINTER(EDFFILE), c_uint]
edf_get_sample_close_to_time.restype = POINTER(ALLF_DATA)

edf_get_element_count = edfapi.edf_get_element_count
edf_get_element_count.argtypes = [POINTER(EDFFILE)]
edf_get_element_count.restype = c_uint

edf_get_revision = edfapi.edf_get_revision
edf_get_revision.argtypes = [POINTER(EDFFILE)]
edf_get_revision.restype = c_int

edf_get_eyelink_revision = edfapi.edf_get_eyelink_revision
edf_get_eyelink_revision.argtypes = [POINTER(EDFFILE)]
edf_get_eyelink_revision.restype = c_int

edf_set_trial_identifier = edfapi.edf_set_trial_identifier
edf_set_trial_identifier.argtypes = [POINTER(EDFFILE), String, String]
edf_set_trial_identifier.restype = c_int

edf_get_start_trial_identifier = edfapi.edf_get_start_trial_identifier
edf_get_start_trial_identifier.argtypes = [POINTER(EDFFILE)]
if sizeof(c_int) == sizeof(c_void_p):
    edf_get_start_trial_identifier.restype = ReturnString
else:
    edf_get_start_trial_identifier.restype = String
    edf_get_start_trial_identifier.errcheck = ReturnString

edf_get_end_trial_identifier = edfapi.edf_get_end_trial_identifier
edf_get_end_trial_identifier.argtypes = [POINTER(EDFFILE)]
if sizeof(c_int) == sizeof(c_void_p):
    edf_get_end_trial_identifier.restype = ReturnString
else:
    edf_get_end_trial_identifier.restype = String
    edf_get_end_trial_identifier.errcheck = ReturnString

edf_get_trial_count = edfapi.edf_get_trial_count
edf_get_trial_count.argtypes = [POINTER(EDFFILE)]
edf_get_trial_count.restype = c_int

edf_jump_to_trial = edfapi.edf_jump_to_trial
edf_jump_to_trial.argtypes = [POINTER(EDFFILE), c_int]
edf_jump_to_trial.restype = c_int


class TRIAL(Structure):
    pass

TRIAL.__slots__ = ['rec', 'duration', 'starttime', 'endtime']
TRIAL._fields_ = [('rec', POINTER(RECORDINGS)), ('duration', c_uint),
                  ('starttime', c_uint), ('endtime', c_uint)]


edf_get_trial_header = edfapi.edf_get_trial_header
edf_get_trial_header.argtypes = [POINTER(EDFFILE), POINTER(TRIAL)]
edf_get_trial_header.restype = c_int

edf_goto_previous_trial = edfapi.edf_goto_previous_trial
edf_goto_previous_trial.argtypes = [POINTER(EDFFILE)]
edf_goto_previous_trial.restype = c_int

edf_goto_next_trial = edfapi.edf_goto_next_trial
edf_goto_next_trial.argtypes = [POINTER(EDFFILE)]
edf_goto_next_trial.restype = c_int

edf_goto_trial_with_start_time = edfapi.edf_goto_trial_with_start_time
edf_goto_trial_with_start_time.argtypes = [POINTER(EDFFILE), c_uint]
edf_goto_trial_with_start_time.restype = c_int

edf_goto_trial_with_end_time = edfapi.edf_goto_trial_with_end_time
edf_goto_trial_with_end_time.argtypes = [POINTER(EDFFILE), c_uint]
edf_goto_trial_with_end_time.restype = c_int

edf_set_bookmark = edfapi.edf_set_bookmark
edf_set_bookmark.argtypes = [POINTER(EDFFILE), POINTER(BOOKMARK)]
edf_set_bookmark.restype = c_int

edf_free_bookmark = edfapi.edf_free_bookmark
edf_free_bookmark.argtypes = [POINTER(EDFFILE), POINTER(BOOKMARK)]
edf_free_bookmark.restype = c_int

edf_goto_bookmark = edfapi.edf_goto_bookmark
edf_goto_bookmark.argtypes = [POINTER(EDFFILE), POINTER(BOOKMARK)]
edf_goto_bookmark.restype = c_int

goto_next_bookmark = edfapi.edf_goto_next_bookmark
goto_next_bookmark.argtypes = [POINTER(EDFFILE)]
goto_next_bookmark.restype = c_int

edf_goto_previous_bookmark = edfapi.edf_goto_previous_bookmark
edf_goto_previous_bookmark.argtypes = [POINTER(EDFFILE)]
edf_goto_previous_bookmark.restype = c_int

edf_get_version = edfapi.edf_get_version
edf_get_version.argtypes = []
if sizeof(c_int) == sizeof(c_void_p):
    edf_get_version.restype = ReturnString
else:
    edf_get_version.restype = String
    edf_get_version.errcheck = ReturnString

edf_get_event = edfapi.edf_get_event
edf_get_event.argtypes = [POINTER(ALLF_DATA)]
edf_get_event.restype = POINTER(FEVENT)

edf_get_sample = edfapi.edf_get_sample
edf_get_sample.argtypes = [POINTER(ALLF_DATA)]
edf_get_sample.restype = POINTER(FSAMPLE)

edf_get_recording = edfapi.edf_get_recording
edf_get_recording.argtypes = [POINTER(ALLF_DATA)]
edf_get_recording.restype = POINTER(RECORDINGS)

edf_get_uncorrected_raw_pupil = edfapi.edf_get_uncorrected_raw_pupil
edf_get_uncorrected_raw_pupil.argtypes = [POINTER(EDFFILE), POINTER(FSAMPLE),
                                          c_int, POINTER(c_float)]
edf_get_uncorrected_raw_pupil.restype = None

edf_get_uncorrected_raw_cr = edfapi.edf_get_uncorrected_raw_cr
edf_get_uncorrected_raw_cr.argtypes = [POINTER(EDFFILE), POINTER(FSAMPLE),
                                       c_int, POINTER(c_float)]
edf_get_uncorrected_raw_cr.restype = None

edf_get_uncorrected_pupil_area = edfapi.edf_get_uncorrected_pupil_area
edf_get_uncorrected_pupil_area.argtypes = [POINTER(EDFFILE), POINTER(FSAMPLE),
                                           c_int]
edf_get_uncorrected_pupil_area.restype = c_uint

edf_get_uncorrected_cr_area = edfapi.edf_get_uncorrected_cr_area
edf_get_uncorrected_cr_area.argtypes = [POINTER(EDFFILE), POINTER(FSAMPLE),
                                        c_int]
edf_get_uncorrected_cr_area.restype = c_uint

edf_get_pupil_dimension = edfapi.edf_get_pupil_dimension
edf_get_pupil_dimension.argtypes = [POINTER(EDFFILE), POINTER(FSAMPLE), c_int,
                                    POINTER(c_uint)]
edf_get_pupil_dimension.restype = None

edf_get_cr_dimension = edfapi.edf_get_cr_dimension
edf_get_cr_dimension.argtypes = [POINTER(EDFFILE), POINTER(FSAMPLE),
                                 POINTER(c_uint)]
edf_get_cr_dimension.restype = None

edf_get_window_position = edfapi.edf_get_window_position
edf_get_window_position.argtypes = [POINTER(EDFFILE), POINTER(FSAMPLE),
                                    POINTER(c_uint)]
edf_get_window_position.restype = None

edf_get_pupil_cr = edfapi.edf_get_pupil_cr
edf_get_pupil_cr.argtypes = [POINTER(EDFFILE), POINTER(FSAMPLE), c_int,
                             POINTER(c_float)]
edf_get_pupil_cr.restype = None

edf_get_uncorrected_cr2_area = edfapi.edf_get_uncorrected_cr2_area
edf_get_uncorrected_cr2_area.argtypes = [POINTER(EDFFILE), POINTER(FSAMPLE),
                                         c_int]
edf_get_uncorrected_cr2_area.restype = c_uint

edf_get_uncorrected_raw_cr2 = edfapi.edf_get_uncorrected_raw_cr2
edf_get_uncorrected_raw_cr2.argtypes = [POINTER(EDFFILE), POINTER(FSAMPLE),
                                        c_int, POINTER(c_float)]
edf_get_uncorrected_raw_cr2.restype = None

edf_set_log_function = edfapi.edf_set_log_function
edf_set_log_function.argtypes = [CFUNCTYPE(UNCHECKED(None), String)]
edf_set_log_function.restype = None
"""
